unit SCont;

interface

uses
  SUtil;

type
  PPointerList = ^TPointerList;
  TPointerList = array[0..MaxInt div SizeOf(Pointer)-1] of Pointer;

  TSListSortCompare = function (Item1, Item2: Pointer): Integer;

  TSList = class(TObject)
  private
    FList: PPointerList;
    FCount: Integer;
    FCapacity: Integer;
  protected
    function Get(Index: Integer): Pointer;
    procedure Grow; virtual;
    procedure Put(Index: Integer; Item: Pointer);
    procedure SetCapacity(NewCapacity: Integer);
  public
    destructor Destroy; override;
    function Add(Item: Pointer): Integer;
    procedure Clear;
    procedure Insert(Index: Integer; Item: Pointer);
    procedure Delete(Index: Integer);
    property Count: Integer read FCount;
    property Items[Index: Integer]: Pointer read Get write Put; default;
    procedure Sort(Compare: TSListSortCompare);
  end;

  TSStrList = class(TObject)
  private
    FList: TSList;
    function Get(Index: Integer): string;
    procedure Put(Index: Integer; const S: string);
    function GetCount: Integer;
  public
    constructor Create;
    destructor Destroy; override;
    function Add(const S: string): Integer;
    procedure Clear;
    property Count: Integer read GetCount;
    property Strings[Index: Integer]: string read Get write Put; default;
    procedure Sort;
  end;

  TSObjList = class(TSList)
  private
    function Get(Index: Integer): TObject;
  public
    destructor Destroy; override;
    procedure Clear;
    procedure Delete(Index: Integer);
    procedure Extract(Index: Integer);
    property Items[Index: Integer]: TObject read Get; default;
  end;


  PSStrDictItem = ^TSStrDictItem;
  TSStrDictItem = record
    Next: PSStrDictItem;
    Key: string;
    Value: string;
  end;
  
  PPSStrDictItemArray = ^TPSStrDictItemArray;
  TPSStrDictItemArray = array[0..0] of PSStrDictItem;

  TSStrDict = class(TObject)
  private
    FHashSize: Cardinal;
    FItems: PPSStrDictItemArray;
    FEachItem: PSStrDictItem;
    FEachIdx: Cardinal;
    function Lookup(const Key: string): PSStrDictItem;
    function Hash(const Key: string): Cardinal;
    procedure Put(const Key: string; const Value: string);
    function Get(const Key: string): string;
  public
    constructor Create(HashSize: Integer);
    destructor Destroy; override;
    function HasKey(const Key: string): Boolean;
    function EachKeys(var Key: string): Boolean;
    property Items[const Key: string]: string read Get write Put; default;
  end;

implementation

{ TSList }

destructor TSList.Destroy;
begin
  Clear;
end;

function TSList.Add(Item: Pointer): Integer;
begin
  Result := FCount;
  if Result = FCapacity then
    Grow;
  FList[Result] := Item;
  Inc(FCount);
end;

procedure TSList.Insert(Index: Integer; Item: Pointer);
begin
  if (Index < 0) or (Index > FCount) then
    Exit;
  if FCount = FCapacity then
    Grow;
  if Index < FCount then
    Move(FList[Index], FList[Index+1], (FCount-Index)*SizeOf(Pointer));
  FList[Index] := Item;
  Inc(FCount);
end;

procedure TSList.Clear;
begin
  FCount := 0;
  SetCapacity(0);
end;

procedure TSList.Delete(Index: Integer);
begin
  if (Index < 0) or (Index >= FCount) then
    Exit;
  Dec(FCount);
  if Index < FCount then
    Move(FList[Index + 1], FList[Index], (FCount - Index) * SizeOf(Pointer));
end;

function TSList.Get(Index: Integer): Pointer;
begin
  if (Index < 0) or (Index >= FCount) then
    Result := nil
  else
    Result := FList[Index];
end;

procedure TSList.Grow;
var
  d: Integer;
begin
  if FCapacity > 64 then
    d := FCapacity div 4
  else if FCapacity > 8 then
    d := 16
  else
    d := 4;
  SetCapacity(FCapacity + d);
end;

procedure TSList.Put(Index: Integer; Item: Pointer);
begin
  if (Index < 0) or (Index >= FCount) then
    Exit;
  FList[Index] := Item;
end;

procedure TSList.SetCapacity(NewCapacity: Integer);
begin
  if (NewCapacity < FCount) or (NewCapacity > (High(TPointerList)+1)) then
    Exit;
  if NewCapacity <> FCapacity then
  begin
    ReallocMem(FList, NewCapacity * SizeOf(Pointer));
    FCapacity := NewCapacity;
  end;
end;

procedure QuickSort(SortList: PPointerList; L, R: Integer;
  SCompare: TSListSortCompare);
var
  i, j: Integer;
  p, t: Pointer;
begin
  repeat
    i := L;
    j := R;
    p := SortList[(L + R) shr 1];
    repeat
      while SCompare(SortList[i], p) < 0 do 
        Inc(i);
      while SCompare(SortList[j], p) > 0 do
        Dec(j);
      if i <= j then
      begin
        t := SortList[i];
        SortList[i] := SortList[j];
        SortList[j] := t;
        Inc(i);
        Dec(j);
      end;
    until i > j;
    if L < j then 
      QuickSort(SortList, L, j, SCompare);
    L := i;
  until i >= R;
end;

procedure TSList.Sort(Compare: TSListSortCompare);
begin
  if (FList <> nil) and (FCount > 0) then
    QuickSort(FList, 0, FCount - 1, Compare);
end;


{ TSStrList }

constructor TSStrList.Create;
begin
  FList := TSList.Create;
end;

destructor TSStrList.Destroy;
begin
  Clear;
  FList.Free;
  inherited Destroy;
end;

procedure TSStrList.Clear;
var
  i: Integer;
begin
  for i := 0 to FList.Count - 1 do
    Dispose(PString(FList[i]));
  FList.Clear;
end;

function TSStrList.Get(Index: Integer): string;
begin
  Result := PString(FList[Index])^;
end;

procedure TSStrList.Put(Index: Integer; const S: string);
begin
  PString(FList[Index])^ := S;
end;

function TSStrList.GetCount: Integer;
begin
  Result := FList.Count;
end;

function TSStrList.Add(const S: string): Integer;
var
  p: PString;
begin
  New(p);
  p^ := S;
  Result := FList.Add(p);
end;

function SStrListCompare(Item1, Item2: Pointer): Integer;
begin
  Result := StrColl(PString(Item1)^, PString(Item2)^);
end;

procedure TSStrList.Sort;
begin
  FList.Sort(SStrListCompare);
end;


{ TSObjList }

destructor TSObjList.Destroy;
begin
  Clear;
  inherited Destroy;
end;

procedure TSObjList.Clear;
var
  i: Integer;
begin
  for i := 0 to Count - 1 do
    Items[i].Free;
  inherited Clear;
end;

procedure TSObjList.Extract(Index: Integer);
begin
  inherited Delete(Index);
end;

procedure TSObjList.Delete(Index: Integer);
begin
  Items[Index].Free;
  Extract(Index);
end;

function TSObjList.Get(Index: Integer): TObject;
begin
  Result := TObject(inherited Get(Index));
end;


{ TSStrDict }

constructor TSStrDict.Create(HashSize: Integer);
begin
  FHashSize := HashSize;
  GetMem(FItems, SizeOf(PSStrDictItem) * FHashSize);
  FillChar(FItems^, SizeOf(PSStrDictItem) * FHashSize, 0);
end;

destructor TSStrDict.Destroy;
var
  i: Integer;
  item: PSStrDictItem;
begin
  for i := 0 to FHashSize-1 do
    while FItems[i] <> nil do
    begin
      item := FItems[i];
      FItems[i] := item.Next;
      Dispose(item);
    end;
  FreeMem(FItems);
end;

function TSStrDict.Hash(const Key: string): Cardinal;
var
  p: PChar;
begin
  Result := 0;
  p := PChar(Key);
  while p^ <> #0 do
  begin
    Result := Byte(p^) + 31 * Result;
    Inc(p);
  end;
  Result := Result mod FHashSize;
end;

function TSStrDict.Lookup(const Key: string): PSStrDictItem;
var
  item: PSStrDictItem;
begin
  item := FItems[Hash(Key)];
  while item <> nil do
  begin
    if item.Key = Key then
    begin
      Result := item;
      Exit;
    end;
    item := item.Next;
  end;
  Result := nil;
end;

procedure TSStrDict.Put(const Key: string; const Value: string);
var
  item: PSStrDictItem;
  h: Integer;
begin
  item := Lookup(Key);
  if Item = nil then
  begin
    New(item);
    item.Key := Key;
    h := Hash(Key);
    item.Next := FItems[h];
    FItems[h] := item;
  end;
  item.Value := Value;
end;

function TSStrDict.Get(const Key: string): string;
var
  item: PSStrDictItem;
begin
  item := Lookup(Key);
  if item = nil then
    Result := ''
  else
    Result := item.Value;
end;

function TSStrDict.HasKey(const Key: string): Boolean;
begin
  Result := Lookup(Key) <> nil;
end;

function TSStrDict.EachKeys(var Key: string): Boolean;
begin
  while FEachItem = nil do
  begin
    if FEachIdx < FHashSize then
    begin
      FEachItem := FItems[FEachIdx];
      Inc(FEachIdx);
    end
    else begin
      FEachIdx := 0;
      FEachItem := nil;
      Result := False;
      Exit;
    end;
  end;

  Key := FEachItem.Key;
  FEachItem := FEachItem.Next;
  Result := True;
end;

end.
